import argparse
import numpy as np

from sklearn.decomposition import DictionaryLearning
import src.utils as utils
import src.archive.dictionary_learning as dict_learning


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument("--data", type=str)
    parser.add_argument("--padded", action="store_true")
    parser.add_argument("--code_sparsity", type=int)
    parser.add_argument("--dict_learning_method", choices=["sklearn", "MOD"])
    parser.add_argument("--results_dir", type=str)
    parser.add_argument("--seed", type=int, default=42)
    parser.add_argument("--no_padding", action="store_true", help="set true to use MEG matrix with 204 columns no pad")

    # Params for MOD
    parser.add_argument("--num_iter", type=int)

    # BP model
    parser.add_argument("--bp_depth", type=int, default=None)
    parser.add_argument("--bp_perm", choices=["identity"], default=None)
    parser.add_argument("--bp_tree", choices=["comb"], default=None)

    # Global optimization
    parser.add_argument("--optimizer", choices=["SGD", "SGDMomentum", "Adam"])
    parser.add_argument("--optim_lr", type=float)
    parser.add_argument("--optim_iter", type=int)
    parser.add_argument("--optim_freq", type=int)
    parser.add_argument("--cuda", type=int)

    args = parser.parse_args()

    meg_matrix = np.load(args.data)
    print("Shape of MEG matrix: ", meg_matrix.shape)
    meg_padded = None
    if args.padded:
        meg_padded = utils.add_padding(meg_matrix)

    if args.dict_learning_method == "sklearn":
        dict_learner = DictionaryLearning(
            transform_algorithm='omp',
            transform_n_nonzero_coefs=args.code_sparsity,
            max_iter=5000
        )
        if meg_padded is not None:
            code = dict_learner.fit_transform(meg_padded)
            meg_padded_approx = code @ dict_learner.components_
            meg_approx = meg_padded_approx[:meg_matrix.shape[0], :meg_matrix.shape[1]]
            loss = np.linalg.norm(meg_matrix - meg_approx) / np.linalg.norm(meg_matrix)
            print("Relative error: ", loss)

    elif args.dict_learning_method == "MOD":
        method = dict_learning.MethodOfOptimalDirections(
            meg_matrix, args.results_dir, args.seed, code_init="random", use_padding=not args.no_padding,
            num_iter=args.num_iter, coder_method="omp", code_sparsity=args.code_sparsity,
            bp_depth=args.bp_depth, bp_perm=args.bp_perm, bp_tree=args.bp_tree,
            optimizer=args.optimizer, optim_lr=args.optim_lr, optim_iter=args.optim_iter, optim_freq=args.optim_freq,
            cuda=args.cuda
        )
        factors = method.factorize()
        print(method.stats.stats["loss"])

    elif args.dict_learning_method == "PALM":
        pass
        # TODO

    elif args.dict_learning_method == "KSVD":
        pass
        # TODO