from pyfaust.fact import hierarchical
from pyfaust.factparams import StoppingCriterion, ParamsHierarchical
from pyfaust.proj import splin, sp
import numpy as np
from numpy.linalg import norm

import argparse
import src.utils as utils
from pathlib import Path

import time

if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument("--code_sparsity", type=int)
    parser.add_argument("--J", type=int)
    parser.add_argument("--num_its", type=int)
    parser.add_argument("--results_dir", type=str)
    parser.add_argument("--gpu", action="store_true")
    parser.add_argument("--float", action="store_true")
    parser.add_argument("--data", default=None)
    args = parser.parse_args()

    print(f"======== J = {args.J}, code_sparsity = {args.code_sparsity}, num_its = {args.num_its}")
    dir_path = Path(args.results_dir)
    dir_path.mkdir(exist_ok=True, parents=True)
    name = f"code_sparsity_{args.code_sparsity}-J_{args.J}-num_its_{args.num_its}"
    stats = utils.Stats(dir_path / f"{name}.pkl", ["code_sparsity", "J", "num_its", "error", "time"])

    # load the MEG matrix
    if args.data is None:
        from pyfaust.demo import get_data_dirpath
        from scipy.io import loadmat
        d = loadmat(get_data_dirpath()+'/matrix_MEG.mat')
        MEG = d['matrix']
    else:
        MEG = np.load(args.data)
    if args.float:
        MEG = MEG.astype("float32")
    n, m = MEG.shape
    print("MEG dims m,n=", m, n)

    # compute m_pad
    m_pad = 256

    # projectors list for factors and residuals
    S_projs = []
    R_projs = []

    # TODO: change J
    # J = 8 # number of factors

    # S1 sparsity
    # k = m / 2 # TODO: change
    S1_proj = splin((n, m), int(args.code_sparsity))
    S_projs += [S1_proj]

    # sparsity for  S_j {2,...J} factors
    s_j = np.empty((args.J - 1,))
    # TODO: set s_j
    s_j[:] = 2 * m_pad  # TODO: change as you need, of course can be different one from each other (s_j[i] != s_j[j])
    for j in range(1, args.J - 1):
        S_j_proj = sp((m, m), int(s_j[j-2]))
        S_projs += [S_j_proj]

    # residual sparsity parameters (as in the paper)
    rho = .8
    if args.J == 3:
        s = 128 ** 2 + 128 * (204 % 128)
    elif args.J == 4:
        s = 3 * 64 ** 2 + 64 * (204 % 64)
    else:
        s = 0
        # P = 1.4 * m ** 2
        assert NotImplementedError
    # residual projectors
    for j in range(args.J - 1):
        R_projs += [sp((m, m), int(s * rho**(j - args.J + 2)))]

    # set the number of PALM4MSA iterations for both split and global optimizations
    #TODO: change (num_its=30 ?)
    split_stop_crit = StoppingCriterion(num_its=args.num_its)
    glob_stop_crit = StoppingCriterion(num_its=args.num_its)
    # if args.J == 3:
    #     num_its = 8000 // 9
    # elif args.J == 4:
    #     num_its = 8000 // 15
    # else:
    #     num_its = 0
    #     assert NotImplementedError

    # instantiate the ParamsHierarchical with this set of parameters
    p = ParamsHierarchical(S_projs, R_projs, split_stop_crit, glob_stop_crit)
    p.factor_format = 'dense'

    # launch the hierarchical factorization
    begin = time.time()
    F = hierarchical(MEG, p, backend=2020, on_gpu=args.gpu)
    total_time = time.time() - begin

    error = norm(F.toarray() - MEG) / norm(MEG)
    # print(F)
    # print(MEG)
    # MEG_f = Faust([MEG], dev='cpu')
    # F_cpu = F.clone('cpu')
    # print(F_cpu - MEG_f)
    # error = (F - MEG).norm() / MEG.norm()
    print("error:", error)
    print("time:", total_time)
    stats.update([args.code_sparsity, args.J, args.num_its, error, total_time])
