import scipy.linalg
from src import tree
import src.factorization as fact
import src.utils as utils
import matplotlib.pyplot as plt
import scipy.fftpack
import numpy as np

from src.utils import matrix_product, best_perm, bit_reversal_permutation_matrix


def alternate_fac(A, support, noise_std, num_iter, tree_type, max_depth=-1):
    assert A.shape[0] == A.shape[1]
    perm = np.zeros(A.shape)
    perm[np.arange(A.shape[0]), np.random.permutation(A.shape[0])] = 1
    losses = []
    if tree_type == "comb":
        generate_partial_tree = tree.generate_partial_comb_tree
        generate_tree = tree.generate_comb_tree
    else:
        assert tree_type == "balanced"
        generate_partial_tree = tree.generate_partial_balanced_tree
        generate_tree = tree.generate_balanced_tree
    for i in range(num_iter):
        if max_depth >= 0:
            root = generate_partial_tree(0, len(support), len(support), 0, max_depth)
        else:
            root = generate_tree(0, len(support), len(support))
        result = fact.tree_hierarchical_factorization(root, A @ np.transpose(perm))
        B = matrix_product(result)
        if np.iscomplexobj(B):
            N = utils.generate_matrix_noise(A.shape, 0, noise_std) + 1j * utils.generate_matrix_noise(A.shape, 0, noise_std)
        else:
            N = utils.generate_matrix_noise(A.shape, 0, noise_std)
        perm = find_best_projection_matrix_in_BP(A, B + N)
        loss = np.linalg.norm(A - B @ perm) / (np.linalg.norm(A) + 1e-15)
        print("Iteration ", i + 1, ": ", loss)
        losses.append(loss)
    return losses


def find_best_projection_matrix_in_BP(A, B):
    new_perm = best_perm(A, B)
    perm = np.zeros(A.shape)
    perm[np.arange(A.shape[0]), new_perm] = 1
    return np.transpose(perm)


if __name__ == '__main__':
    n = 4
    support = utils.support_DFT(n)

    # B = scipy.linalg.hadamard(2 ** n)
    B = utils.generate_random_matrix(support, main_m=0., main_std=1)
    true_perm = perm = np.zeros(B.shape)
    perm[np.arange(B.shape[0]), np.random.permutation(B.shape[0])] = 1
    A = B @ true_perm

    # A = scipy.linalg.dft(2 ** n)

    # perm = find_best_projection_matrix_in_BP(A, B)
    # print(perm == true_perm)
    # print(np.linalg.pinv(perm) == np.transpose(perm))

    alternate_fac(A, support, noise_std=0., num_iter=10, max_depth=1, tree_type="balanced")
