import numpy as np
#from pyfaust import *
import matplotlib.pyplot as plt

from src.archive.palm import projection, palm_factorization_two_factors


def gradient_descent(A, B, C, lr = 2e-3, max_iter = 50):
  supportA = np.where(A!=0, 1, 0)
  supportB = np.where(B!=0, 1, 0)

  for _ in range(max_iter):
    diff = A @ B - C
    A = np.multiply(A - lr * np.dot(diff, B.T), supportA) 
    #diff = A @ B - C
    B = np.multiply(B - lr * np.dot(A.T, diff), supportB)
  return A, B

def gradient_descent_modified(A, B, M, k, max_iter = 50):
    for i in range(max_iter):
        adaptlr = (1 + 1e-3) * np.linalg.norm(B, ord = 2) ** 2
        A = A - (1 / adaptlr) * (A @ B - M) @ B.T
        suppA = projection(A.T, k).T
        A = A * suppA 
    
        adaptlr = (1 + 1e-3) * np.linalg.norm(A, ord = 2) ** 2
        B = B - (1 / adaptlr) * np.dot(A.T, A @ B - M)
    
    return A,B


def MEG_fac(M, k, lr = 1e-3, max_iter = 10000):
    """
    A = np.zeros((M.shape[0], M.shape[1])) 
    B = np.eye(M.shape[1], M.shape[1]) * 1e-1  #np.random.randn(M.shape[1], M.shape[1]) * 1e-1#
    """
    """
    A = np.zeros((M.shape[0], M.shape[1])) 
    B = np.random.randn(M.shape[1], M.shape[1]) #
    """
    """
    B = np.zeros((M.shape[1], M.shape[1]))
    A = np.eye(M.shape[0], M.shape[1])
    """
    
    B = np.eye(M.shape[1], M.shape[1])
    A = np.eye(M.shape[0], M.shape[1])
    
    
    """
    for i in range(max_iter):
        if i % 50 == 0 and i < 500:
            print("Support change")
            suppA = np.where(A!=0, 1, 0)
            temp = HTP(B.T, M.T, k).T
            suppAnew = np.where(temp!=0, 1, 0)
            if i ==0 or np.sum(np.abs(suppA - suppAnew)) > 1e-3:
                print("Support change complete")
                A = temp
            else:
                print("No support change")
        #A, B = gradient_descent(A, B, M, lr = lr, max_iter = 20)
        A, B = gradient_descent_modified(A,B,M,k, max_iter = 2)
        print("Iteration ", i, ": ", np.linalg.norm(np.dot(A,B) - M) / np.linalg.norm(M))
    """
    
    lamda = 1.0
    
    A, B = initilization(M * 1.0)
    lamda = np.linalg.norm(A) * np.linalg.norm(B)
    A = A / np.linalg.norm(A)
    B = B / np.linalg.norm(B)
    
    for i in range(max_iter):
        adaptlr = (1 + 1e-3) * lamda * lamda * np.linalg.norm(B, ord = 2) ** 2
        A = A - (1 / adaptlr) * lamda * (lamda * A @ B - M) @ B.T
        suppA = projection(A.T, k).T
        A = A * suppA 
        A = A / np.linalg.norm(A)
    
        adaptlr = (1 + 1e-3) * lamda * lamda * np.linalg.norm(A, ord = 2) ** 2
        B = B - (1 / adaptlr) * np.dot(lamda * A.T, lamda * A @ B - M)
        B = B / np.linalg.norm(B)
        
        #B = B - lr * np.dot(A.T, A @ B - M)
        r = np.dot(A,B)
        lamda = np.trace(np.dot(M.T, r)) / np.trace(np.dot(r.T,r))
        #supportA = np.where(np.absolute(A) > 1e-15, 1, 0)
        #supportB = np.where(np.absolute(B) > 1e-15, 1, 0)
        #print(np.sum(supportA, axis = 1), np.sum(supportB, axis = 0))
        #print("Iteration ", i, ": ", np.linalg.norm(lamda * np.dot(A,B) - M) / np.linalg.norm(M))
        supportA = np.where(np.absolute(A) > 1e-17, 1, 0)
        supportB = np.where(np.absolute(B) > 1e-17, 1, 0)
        #print(np.sum(supportA), np.sum(supportB))
    """
    #this piece of code works well
    for i in range(max_iter):
        A = A - lr * (A @ B - M) @ B.T
        suppA = projection(A.T,k).T
        A = A * suppA 
        B = B - lr * np.dot(A.T, A @ B - M)
        print("Iteration ", i, ": ", np.linalg.norm(np.dot(A,B) - M) / np.linalg.norm(M))
    """
    """
    lamda = 1
    B = B / np.linalg.norm(B)
    for i in range(max_iter):
        A = A - lr * lamda * (lamda * A @ B - M) @ B.T
        suppA = projection(A.T,k).T
        A = A * suppA 
        A = A / np.linalg.norm(A)
        B = B - lr * np.dot(lamda * A.T, lamda * A @ B - M)
        B = B / np.linalg.norm(B)
        r = np.dot(A,B)
        lamda = np.trace(np.dot(M.T, r)) / np.trace(np.dot(r.T,r))
        print("Iteration ", i, ": ", np.linalg.norm(lamda * np.dot(A,B) - M) / np.linalg.norm(M))
    """
    return A,B


def initilization(matrix):
    comparison = np.zeros((matrix.shape[0], matrix.shape[0]))
    A = np.zeros(matrix.shape)
    B = np.zeros((matrix.shape[1], matrix.shape[1]))
    last_i = -1
    last_j = -1
    
    for k in range(matrix.shape[1]):
        if k == 0:    
            for i in range(matrix.shape[0]):
                for j in range(i + 1, matrix.shape[0]):
                    u, s, vh = np.linalg.svd(matrix[[i,j]])
                    comparison[i,j] = -s[0] * s[0]
        else:
            for i in range(matrix.shape[0]):
                if i == last_i:
                    continue
                u, s, vh = np.linalg.svd(matrix[[i,last_i]])
                comparison[min(i, last_i),max(i, last_i)] = -s[0] * s[0]
            
            for i in range(matrix.shape[0]):
                if i == last_j:
                    continue
                u, s, vh = np.linalg.svd(matrix[[i,last_j]])
                comparison[min(i, last_j),max(i, last_j)] = -s[0] * s[0]
        index = np.argmin(comparison)
        last_i, last_j = np.unravel_index(np.argmin(comparison), comparison.shape)
        u, s, vh = np.linalg.svd(matrix[[last_i, last_j]])
        matrix[[last_i, last_j]] -= s[0] * np.outer(u[:,0], vh[0])
        print(np.linalg.norm(matrix))
        A[[last_i, last_j], [k] * 2] = u[:,0] * s[0]
        B[k] = vh[0]
    
    return A,B


if __name__ == '__main__':
    MEG = np.load('MEG.npy')
    print(MEG.shape)
    #MEG = np.load('gaussian1.npy')
    #MEG = Faust(MEG).T.normalize(2).toarray().T
    #print(np.linalg.norm(MEG))
    #A,B = MEG_fac(MEG, 5, max_iter = 2000)
    #_ = palm_factorization_two_factors(MEG, 100, 25, max_iter = 100)
    #_ = palm_factorization_two_factors(MEG, 100, 25, max_iter = 100, flag = False)
    #MEG = np.load('gaussian.npy')
    evol1 = palm_factorization_two_factors(MEG, 25, 25, max_iter = 1000, flag = False)
    evol2 = palm_factorization_two_factors(MEG, 25, 25, max_iter = 1000, flag = True)
    
    plt.figure(figsize = (10,5))
    plt.plot(evol1, label = 'PALM')
    plt.plot(evol2, label = 'BHTP')
    plt.xlabel('iterations')
    plt.ylabel('error')
    plt.legend(prop={'size' : 15})
    plt.savefig('evol_HTP.png', dpi = 100, bbox = 'tight')
    
    #A, B = MEG_fac(MEG, 2,max_iter = 5000)
    #print(A)
    #A, B = initilization(MEG * 1.0)
    #print(np.linalg.norm(np.dot(A,B) - MEG))
    #print(np.linalg.norm(np.dot(A,B) - MEG) / np.linalg.norm(MEG))
    
    
