#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import torch
import numpy as np
from src.archive.masked_linear import CustomizedLinear

# define mask matrix to customize linear
mask = torch.tensor(
  [[1, 0, 1],
   [0, 1, 0],
   [1, 0, 1],
   [1, 0, 1],]
  )

# define size of layers.
# this architect is [INPUT, HIDDEN(masked(customized) linear), OUTPUT]-layers.
Dim_INPUT  = mask.shape[0]
Dim_HIDDEN = mask.shape[1]
Dim_OUTPUT = 1

# create randomly input:x, output:y as train dataset.
batch = 1
x = torch.randn(batch, Dim_INPUT)
y = torch.randn(batch, Dim_OUTPUT)

# pipe as model
model = torch.nn.Sequential(
        CustomizedLinear(mask, bias=None), # dimmentions is set from mask.size
        torch.nn.Linear(Dim_HIDDEN, Dim_OUTPUT, bias=False),
        )

# backward pass
print('=== mask matrix ===')
print(mask)
print('===================')
learning_rate = 0.1
for t in range(3):
    # forward
    y_pred = model(x)

    # loss
    loss = (y_pred - y).abs().mean()

    # Zero the gradients before running the backward pass.
    model.zero_grad()

    # Use autograd to compute the backward pass
    loss.backward()

    # Update the weights
    with torch.no_grad():
        for param in model.parameters():
            # mask is also saved in param, but mask.requires_grad=False
            if param.requires_grad:
                param -= learning_rate * param.grad
                # check masked param.grad
                if np.array(param.grad).size == np.array(mask).size:
                    print('--- epoch={}, loss={} ---'.format(t,loss.item()))
                    print('↓↓↓masked weight↓↓↓')
                    print(param.t())
                    print('↓↓↓masked grad of weight↓↓↓')
                    print(param.grad.t())
