import argparse
import scipy.linalg
import src.tree as tree
import src.factorization as fact
import src.utils as utils
import numpy as np
import matplotlib.pyplot as plt


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument("--num_factors", default=5)
    parser.add_argument("--max_depth", default=5)
    parser.add_argument("--noise_std", type=float, default=0.1)

    args = parser.parse_args()

    matrix = scipy.linalg.hadamard(2 ** args.num_factors)
    # Add noise if wanted
    if args.noise_std:
        print(f"Adding noise with std {args.noise_std}")
        noise = utils.generate_matrix_noise(matrix.shape, 0, args.noise_std)
        matrix = matrix + noise

    losses = []
    depths = range(1, args.max_depth + 1)
    for depth in depths:
        partial_tree = tree.generate_partial_comb_tree(0, args.num_factors, args.num_factors, 0, depth)
        print(partial_tree.print_tree())
        result = fact.tree_hierarchical_factorization(partial_tree, matrix)
        loss = utils.error_cal(result, matrix)
        print(f"Partial tree of depth {depth}:", loss)
        losses.append(loss)

    plt.plot(depths, losses)
    plt.show()
