import argparse
import scipy.linalg
import src.tree as tree
import src.factorization as fact
import src.utils as utils
import numpy as np
import matplotlib.pyplot as plt


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument("--num_factors", default=6)
    parser.add_argument("--depth", default=2)
    parser.add_argument("--noise_std", type=float, default=0.1)

    args = parser.parse_args()
    matrix = scipy.linalg.hadamard(2 ** args.num_factors)
    # Add noise if wanted
    if args.noise_std:
        print(f"Adding noise with std {args.noise_std}")
        noise = utils.generate_matrix_noise(matrix.shape, 0, args.noise_std)
        matrix = matrix + noise

    balanced_root = tree.generate_partial_balanced_tree(0, args.num_factors, args.num_factors, 0, args.depth)
    comb_root = tree.generate_partial_comb_tree(0, args.num_factors, args.num_factors, 0, args.depth)

    result = fact.tree_hierarchical_factorization(balanced_root, matrix)
    loss = utils.error_cal(result, matrix)
    print(f"Partial balanced tree of depth {args.depth}:", loss)

    result = fact.tree_hierarchical_factorization(comb_root, matrix)
    loss = utils.error_cal(result, matrix)
    print(f"Partial comb tree of depth {args.depth}:", loss)
