import argparse
import scipy

import numpy as np
import src.utils as utils
import src.factorization as fact

if __name__ == '__main__':
    parser = argparse.ArgumentParser(description="BP projection with fixed P")
    parser.add_argument("--num_factors", type=int, default=10)
    parser.add_argument("--perm", choices=["identity", "bit-reversal"], default="identity")
    parser.add_argument("--matrix", choices=["dft", "hadamard"], default="hadamard")
    parser.add_argument("--tree", choices=["comb", "balanced", "inversed_comb"], default="balanced",
                        help="Default is Tung's baseline. Choosing a tree leads to tree multilayer factorization.")
    args = parser.parse_args()

    if args.matrix == "dft":
        matrix = scipy.linalg.dft(2 ** args.num_factors)
    else:
        assert args.matrix == "hadamard"
        matrix = scipy.linalg.hadamard(2 ** args.num_factors)

    p = utils.get_permutation_matrix(args.num_factors, args.perm)
    projection = fact.project_BP_model_P_fixed(matrix, args.tree, p, max_depth=1)

    print(np.linalg.norm(matrix - projection) / (np.linalg.norm(matrix) + 1e-15))
