import numpy as np
from sklearn.decomposition import SparseCoder
import src.utils as utils
import src.factorization as fact
import time
import argparse

def projection(A, k):
    result = np.zeros(A.shape)
    order = np.argsort(np.absolute(A), axis=0)
    row_index = order[-k:].T.flatten()
    col_index = np.tile(np.arange(A.shape[1]).astype(np.int), (k, 1)).T.flatten()
    result[row_index, col_index] = 1
    return result

def learning_with_butterfly_model(data, col_sparsity, num_iter, tree, perm_name, coder_method, max_depth,
                                  random_permute_MEG):
    if random_permute_MEG:
        print("Permuting MEG rows randomly")
        data = data[np.random.permutation(len(data))]
    #data = np.random.randn(data.shape[0], data.shape[1])
    data_pad = utils.add_padding(data)
    #data_pad = np.random.randn(data_pad.shape[0], data_pad.shape[1])
    code = np.eye(data_pad.shape[0], data_pad.shape[1])#np.random.randn(data_pad.shape[0], data_pad.shape[1])
    num_factors = int(np.log2(data_pad.shape[1]))
    dictionary = np.eye(2 ** num_factors)
    p = None
    if perm_name != "all_8":
        p = utils.get_permutation_matrix(num_factors, perm_name)
        print(p)
    for i in range(num_iter):
        print(f"Iteration {i}")
        """
        print("Pseudo inverse")
        matrix = np.linalg.pinv(code) @ data_pad
        if p is None:
            print(f"BP projection, where P can be any of the fixed 8, and {tree} tree, max_depth={max_depth}")
            product = fact.project_BP_model_8_perm_fixed(matrix, tree, max_depth)
        else:
            print(f"BP projection, with fixed P = {perm_name}, and {tree} tree, max_depth={max_depth}")
            product = fact.project_BP_model_P_fixed(matrix, tree, p, max_depth)
        print("Sparse coding")
        #product = matrix
        norms = np.linalg.norm(product, axis=1)[:, np.newaxis]
        dictionary = product / np.linalg.norm(product, axis=1)[:, np.newaxis]
        coder = SparseCoder(
            dictionary=dictionary,
            transform_n_nonzero_coefs=col_sparsity,
            transform_algorithm=coder_method
        )
        
        code = coder.transform(data_pad)
        """
        for j in range(20):
            begin = time.time()
            adaptlr = (1 + 1e-3) * np.linalg.norm(dictionary, ord = 2) ** 2
            code = code - (1.0 / adaptlr) * (code @ dictionary - data_pad) @ dictionary.T
            supp = projection(code.T, col_sparsity).T
            code = code * supp

            adaptlr = (1 + 1e-3) * np.linalg.norm(code, ord = 2) ** 2
            dictionary = dictionary - (1.0 / adaptlr) * code.T @ (code @ dictionary - data_pad)
            end = time.time()
            #print(end - begin)
            #print(np.linalg.norm(dictionary[:, data.shape[1]:]))
            dictionary[:, data.shape[1]:] = 0
            #print(np.max(np.abs(dictionary)))
            begin = time.time()
            dictionary = fact.project_BP_model_P_fixed(dictionary, tree, p, max_depth)
            end = time.time()
            #print("Proj: ", end - begin)
        loss = utils.relative_error_padded_product(data, code, dictionary)
        print("Loss:", loss)
    return code, dictionary

def parse_args():
    parser = argparse.ArgumentParser(description="MEG factorization")
    parser.add_argument("--k", type=int, default=5)
    parser.add_argument("--num_iter", type=int, default=200)
    parser.add_argument("--max_depth", type=int, default=1)
    parser.add_argument("--tree", type=str, choices=["comb", "balanced", "inverse_comb"], default="comb")
    return parser.parse_args()

if __name__ == "__main__":
    arg = parse_args()
    MEG = np.load("../data/MEG.npy")
    col_sparsity = arg.k
    print(col_sparsity)
    num_iter = arg.num_iter
    tree = arg.tree
    perm_name = "identity"
    coder_method = "omp"
    max_depth = arg.max_depth
    random_permute_MEG = False
    learning_with_butterfly_model(MEG, col_sparsity, num_iter, tree, perm_name, coder_method, max_depth, random_permute_MEG)
