import numpy as np
import src.factorization as fact
import src.utils as utils

import torch

def projection(A, k):
    result = np.zeros(A.shape)
    order = np.argsort(np.absolute(A), axis=0)
    row_index = order[-k:].T.flatten()
    col_index = np.tile(np.arange(A.shape[1]).astype(np.int), (k, 1)).T.flatten()
    result[row_index, col_index] = 1
    return result


def HTP(A, C, k, max_iter = 50, mu = 1e-3, use_proj = False, grad_iter = 10, step = 1e-3):
  x = np.zeros((A.shape[1], C.shape[1]))
  B = np.dot(A.T, A)
  z = np.dot(A.T, C)

  for i in range(max_iter):
    temp = x + mu * (z - np.dot(B, x))
    #print(temp)
    alive = projection(temp, k)

    new_x = np.zeros(x.shape)
    if use_proj:
      for j in range(x.shape[0]):
        support = np.argsort(alive[:, j])[-k:]
        new_x[support, j] = np.dot(np.linalg.pinv(A[:, support]), C[:, j])
    else:
      new_x = x * 1.0
      for j in range(grad_iter):
        new_x = np.multiply(new_x + step * (z - np.dot(B,new_x)), alive)

    x = new_x
  return x


def palm_factorization_two_factors(M, k0, k1, max_iter, random_perm, flag, stats, lr=1e-4):
    """
    A = np.zeros((M.shape[0], M.shape[1]))
    B = np.eye(M.shape[1], M.shape[1]) * 1e-1  #np.random.randn(M.shape[1], M.shape[1]) * 1e-1#
    """
    """
    A = np.random.randn(M.shape[0], M.shape[1])#np.zeros((M.shape[0], M.shape[1])) 
    B = np.random.randn(M.shape[1], M.shape[1]) #
    """
    """
    B = np.zeros((M.shape[1], M.shape[1]))
    A = np.eye(M.shape[0], M.shape[1])
    """
    if random_perm:
        M = M[np.random.permutation(len(M))]

    B = np.eye(M.shape[1], M.shape[1])
    A = np.eye(M.shape[0], M.shape[1])

    evol = []

    bestA = np.zeros(A.shape)
    bestB = np.zeros(B.shape)
    score = 1.0

    for i in range(max_iter):

        # Begin comment
        if i % 50 == 0 and flag:
            print("Doing BHTP step")
            suppA = np.where(A!=0, 1, 0)
            temp = HTP(B.T, M.T, k0, step = 1e-3).T
            suppAnew = np.where(temp!=0, 1, 0)
            if i ==0 or np.sum(np.abs(suppA - suppAnew)) > 1e-3:
                A = temp

            suppB = np.where(B!=0, 1, 0)
            temp = HTP(A, M, k1, step = 1e-4)
            suppBnew = np.where(temp!=0, 1, 0)
            if i ==0 or np.sum(np.abs(suppB - suppBnew)) > 1e-3:
                B = temp
        #end comment

        normA = np.linalg.norm(A)
        normB = np.linalg.norm(B)
        A = A * np.sqrt(normB / normA)
        B = B * np.sqrt(normA / normB)

        supportA = np.where(np.absolute(A) > 1e-17, 1, 0)
        supportB = np.where(np.absolute(B) > 1e-17, 1, 0)

        for loz in range(20):
            adaptlr = (1 + 1e-3) * np.linalg.norm(B, ord = 2) ** 2
            A = A - (1 / adaptlr) * (A @ B - M) @ B.T
            suppA = projection(A.T, k0).T
            A = A * suppA

            adaptlr = (1 + 1e-3) * np.linalg.norm(A, ord = 2) ** 2
            B = B - (1 / adaptlr) * np.dot(A.T, A @ B - M)
            suppB = projection(B, k1)
            B = B * suppB
        rekt = np.linalg.norm(np.dot(A, B) - M) / np.linalg.norm(M)
        evol.append(rekt)
        stats.update([k0, k1, random_perm, flag, i, rekt])
        if score > np.linalg.norm(np.dot(A, B) - M) / np.linalg.norm(M):
            score = np.linalg.norm(np.dot(A, B) - M) / np.linalg.norm(M)
            bestA = A
            bestB = B
        print("Iteration ", i, ": ", np.linalg.norm(np.dot(A, B) - M) / np.linalg.norm(M))
    print(score)

    return evol


def palm_with_BP_factorization_two_factors(data, k0, k1, max_iter, random_perm, flag, stats, lr=1e-4):
    if random_perm:
        data = data[np.random.permutation(len(data))]

    M = utils.add_padding(data)

    B = np.eye(M.shape[1], M.shape[1])
    A = np.eye(M.shape[0], M.shape[1])

    evol = []

    bestA = np.zeros(A.shape)
    bestB = np.zeros(B.shape)
    score = 1.0

    for i in range(max_iter):

        normA = np.linalg.norm(A)
        normB = np.linalg.norm(B)
        A = A * np.sqrt(normB / normA)
        B = B * np.sqrt(normA / normB)

        supportA = np.where(np.absolute(A) > 1e-17, 1, 0)
        supportB = np.where(np.absolute(B) > 1e-17, 1, 0)

        for loz in range(20):
            adaptlr = (1 + 1e-3) * np.linalg.norm(B, ord = 2) ** 2
            A = A - (1 / adaptlr) * (A @ B - M) @ B.T
            suppA = projection(A.T, k0).T
            A = A * suppA

            adaptlr = (1 + 1e-3) * np.linalg.norm(A, ord = 2) ** 2
            B = B - (1 / adaptlr) * np.dot(A.T, A @ B - M)
            # BP Projection !
            B = fact.project_BP_model_P_fixed(B, "comb", np.identity(B.shape[0]), max_depth=1)

        rekt = utils.relative_error_padded_product(data, A, B)
        evol.append(rekt)
        stats.update([k0, k1, random_perm, flag, i, rekt])
        if score > rekt:
            score = rekt
            bestA = A
            bestB = B
        print("Iteration ", i, ": ", rekt)
    print(score)

    return evol


def torch_projection(A, k):
    result = torch.zeros_like(A)
    order = torch.argsort(torch.absolute(A), dim=0)
    row_index = torch.flatten(order[-k:].t())
    col_index = torch.flatten(torch.tile(torch.arange(A.size()[1], dtype=torch.long), (k, 1)).t())
    result[row_index, col_index] = 1
    return result


def torch_relative_error_padded_product(M, *factors, eps=1e-15):
    product_pad = factors[0]
    for i in range(1, len(factors)):
        product_pad = torch.mm(product_pad, factors[i])
    product = product_pad[:M.shape[0], :M.shape[1]]
    return torch.norm(M - product) / (torch.norm(M) + eps)


def torch_palm_with_BP_factorization_two_factors(data, k0, k1, max_iter, random_perm, flag, stats, lr=1e-4):
    if random_perm:
        data = data[np.random.permutation(len(data))]

    print("Converting to cuda tensors")
    M = torch.from_numpy(utils.add_padding(data)).to(torch.double).cuda()
    data = torch.from_numpy(data).cuda()
    B = torch.eye(M.shape[1], M.shape[1]).to(torch.double).cuda()
    A = torch.eye(M.shape[0], M.shape[1]).to(torch.double).cuda()

    evol = []
    score = 1.0
    for i in range(max_iter):
        normA = torch.norm(A)
        normB = torch.norm(B)
        A = A * torch.sqrt(normB / normA)
        B = B * torch.sqrt(normA / normB)

        for loz in range(1):
            adaptlr = (1 + 1e-3) * torch.norm(B) ** 2
            A = A - (1 / adaptlr) * torch.mm((torch.mm(A, B) - M), B.t())
            suppA = torch_projection(A.t(), k0).t()
            A = A * suppA

            adaptlr = (1 + 1e-3) * torch.norm(A) ** 2
            B = B - (1 / adaptlr) * torch.mm(A.t(), torch.mm(A, B) - M)
            # BP Projection !
            B_numpy = fact.project_BP_model_P_fixed(B.cpu().numpy(), "comb", np.identity(B.shape[0]), max_depth=1)
            B = torch.from_numpy(B_numpy).to(torch.double).cuda()

        rekt = torch_relative_error_padded_product(data, A, B)
        evol.append(rekt)
        stats.update([k0, k1, random_perm, flag, i, rekt])
        if score > rekt:
            score = rekt
        print("Iteration ", i, ": ", rekt)
    print(score)

    return evol